import os
import pwd
import random
import string
import subprocess
from charms.apt import purge, queue_install
from charms.reactive import clear_flag, endpoint_from_flag, set_flag, when, when_any, when_not  # noqa: F401
from charmhelpers.core import hookenv, host, unitdata
from charmhelpers.core.templating import render
from charmhelpers.contrib.charmsupport import nrpe


# We actually have two configs for this, though realistically the hyperkitty
# one will only change if the archiver relation changes but it doesn't hurt
# anything to write them both.
def write_config():
    db = unitdata.kv()
    config = hookenv.config()
    config.update(db.getrange('admin_'))
    config.update(db.getrange('db_'))
    config.update(db.getrange('archiver_'))
    render(source="mailman.tmpl",
           target="/etc/mailman3/mailman.cfg",
           owner="root",
           group="list",
           perms=0o640,
           context={'config': config})
    render(source="mailman-hyperkitty.tmpl",
           target="/etc/mailman3/mailman-hyperkitty.cfg",
           owner="root",
           group="list",
           perms=0o640,
           context={'config': config})
    host.service_restart('mailman3')


# Once mailman is installed, we can generate our REST API password
@when('apt.installed.mailman3')
@when_not('mailman.initial_setup')
def initial_setup():
    db = unitdata.kv()
    if not db.get('admin_password'):
        # Generate a relatively random password
        chars = string.ascii_letters + string.digits
        password = ''.join((random.choice(chars)) for x in range(40))
        db.set('admin_password', password)
        # Let's just hard set this for now, we can expose it in the juju config later
        db.set('admin_user', 'restadmin')
        db.set('admin_port', 8001)
        db.set('admin_ip', hookenv.unit_get('private-address'))
    queue_install(['python3-mailman-hyperkitty'])
    set_flag('mailman.initial_setup')


@when('apt.installed.postfix')
@when_not('mailman.postfix.configured')
def configure_postfix():
    config = hookenv.config()
    hookenv.status_set('maintenance', 'configuring postfix for posts to mailing lists')

    myhostname = config.get('hostname', subprocess.getoutput(['hostname', '--fqdn']).strip())
    subprocess.check_output(['postconf', 'myhostname={}'.format(myhostname)])

    aliases_map = config.get('aliases_map')
    if aliases_map:
        host.write_file('/etc/postfix/virtual_alias_maps', aliases_map)
        subprocess.check_output(['postmap', '/etc/postfix/virtual_alias_maps'])
        subprocess.check_output(['postconf', 'virtual_alias_maps=hash:/etc/postfix/virtual_alias_maps'])

    smtpd_tls_cert_file = config.get('ssl_cert', '')
    if not smtpd_tls_cert_file and os.path.exists('/etc/ssl/certs/ssl-cert-snakeoil.pem'):
        smtpd_tls_cert_file = '/etc/ssl/certs/ssl-cert-snakeoil.pem'
    smtpd_tls_key_file = config.get('ssl_key', '')
    if not smtpd_tls_key_file and os.path.exists('/etc/ssl/private/ssl-cert-snakeoil.key'):
        smtpd_tls_key_file = '/etc/ssl/private/ssl-cert-snakeoil.key'
    if smtpd_tls_cert_file and smtpd_tls_key_file:
        subprocess.check_output(['postconf', 'smtpd_use_tls=yes',
                                 'smtpd_tls_cert_file={}'.format(smtpd_tls_cert_file),
                                 'smtpd_tls_key_file={}'.format(smtpd_tls_key_file)])
    else:
        subprocess.check_output(['postconf', 'smtpd_use_tls=no'])
    smtp_tls_policy_maps = config.get('ssl_policy_map', '')
    if smtp_tls_policy_maps:
        subprocess.check_output(['postconf', 'smtp_tls_security_level=may',
                                 'smtp_tls_policy_maps=hash:/etc/postfix/tls_policy'])
        host.write_file('/etc/postfix/tls_policy', smtp_tls_policy_maps)
        subprocess.check_output(['postmap', '/etc/postfix/tls_policy'])

    subprocess.check_output(['postconf', 'smtp_address_preference=ipv4', 'lmtp_address_preference=ipv4',
                             'default_transport=smtp', 'relay_transport=smtp',
                             'smtpd_relay_restrictions=permit_mynetworks defer_unauth_destination'])
    if os.path.exists('/var/lib/mailman3/data/postfix_lmtp'):
        subprocess.check_output(['postconf', 'transport_maps=hash:/var/lib/mailman3/data/postfix_lmtp',
                                 'relay_recipient_maps=hash:/var/lib/mailman3/data/postfix_lmtp',
                                 'local_recipient_maps=hash:/var/lib/mailman3/data/postfix_lmtp'])
        subprocess.check_output(['postmap', '/var/lib/mailman3/data/postfix_lmtp'])
    if os.path.exists('/var/lib/mailman3/data/postfix_domains'):
        subprocess.check_output(['postconf', 'relay_domains=hash:/var/lib/mailman3/data/postfix_domains'])
        subprocess.check_output(['postmap', '/var/lib/mailman3/data/postfix_domains'])

    subprocess.check_output(['postconf', 'smtpd_sender_restrictions=reject_unknown_sender_domain'])

    # HyperKitty expects messages will all have the 'Message-ID' header,
    # unfortunately not all clients do so let's work around it (LP: #1776658)
    subprocess.check_output(['postconf', 'always_add_missing_headers=yes',
                                         'local_header_rewrite_clients=static:all'])

    if config["enable_spamassassin"]:
        configure_spamassassin()
    if config["enable_spf"]:
        configure_postfix_spf()

    configure_recipient_restrictions()

    host.service_reload('postfix')
    # Open the SMTP port
    hookenv.open_port(25)
    set_flag('mailman.postfix.configured')
    hookenv.status_set('active', 'ready')


# Configure smtpd_recipient_restrictions depending on if SPF is enabled
# and if strict hosts checking is enabled or in test mode
def configure_recipient_restrictions():
    config = hookenv.config()

    smtpd_rr = [
        "permit_sasl_authenticated",
        "permit_mynetworks"
    ]

    rr_shc = [
        "reject_non_fqdn_helo_hostname",
        "reject_non_fqdn_recipient",
        "reject_non_fqdn_sender",
        "reject_invalid_helo_hostname",
        "reject_unknown_recipient_domain",
        "reject_unknown_sender_domain"
    ]

    if config["enable_strict_host_checks"] == 'true':
        smtpd_rr.extend(rr_shc)
    elif config["enable_strict_host_checks"] == 'test':
        smtpd_rr.extend(["warn_if_reject " + e for e in rr_shc])

    smtpd_rr.append("reject_unauth_destination")

    if config["enable_spf"]:
        smtpd_rr.append("check_policy_service unix:private/policy-spf")

    # postconf accepts space-separated restrictions, but we also may have
    # whitespace when using eg. warn_if_reject, so let's use commas.
    postconf(key="smtpd_recipient_restrictions", value=",".join(smtpd_rr))


def configure_spamassassin():
    config = hookenv.config()
    default_config = ('ENABLED=0\n'
                      'OPTIONS="--create-prefs --max-children 5 --username spamd --helper-home-dir /var/lib/spamd -s /var/log/spamd.log"\n'
                      'CRON=1\n')
    if config["sa_http_proxy"]:
        default_config += 'http_proxy="' + config["sa_http_proxy"] + '"\n'
        default_config += 'https_proxy="' + config["sa_http_proxy"] + '"\n'

    sa_policy = config["spamassassin_policy"]
    render(source="sa-local.cf.tmpl",
           target="/etc/spamassassin/local.cf",
           owner="root",
           group="root",
           perms=0o444,
           context={'policy': sa_policy})

    host.write_file("/etc/default/spamassassin", default_config)
    if not postconf(key="spamassassin", args=("-F",)):  # Check to see if Spamassassin is already configured.
        postfix_master_cf = os.path.join(hookenv.charm_dir(), "files/spamassassin_master.cf")
        with open(postfix_master_cf) as f:
            append_to_file("/etc/postfix/master.cf", f.read())
    postconf(key="smtp/inet/content_filter", value="spamassassin:dummy", args=("-P",))

    useradd("spamd")
    systemctl("reload", "postfix.service")
    systemctl("enable", "spamassassin.service")
    systemctl("start", "spamassassin.service")
    set_flag('mailman.postfix.spamassassin.configured')


def configure_postfix_spf():
    config = hookenv.config()
    postconf(key="policy-spf_time_limit", value="3600s")

    if not postconf(key="policy-spf", args=("-F",)):  # Check to see if spf is already configured.
        spf_master_cf = os.path.join(hookenv.charm_dir(), "files/spf_master.cf")
        with open(spf_master_cf) as f:
            append_to_file("/etc/postfix/master.cf", f.read())

    host.write_file("/etc/postfix-policyd-spf-python/policyd-spf.conf", config["spf_config"])
    systemctl("reload", "postfix.service")
    set_flag('mailman.postfix.spf.configured')


# HyperKitty's recommended dependencies include, basically, all of Django so
# we go ahead and purge that, we just want the mailman plugin, not the whole
# thing.
@when('apt.installed.python3-mailman-hyperkitty')
@when_not('mailman.django.purged')
def purge_extra_mailman_crap():
    purge('python-django-hyperkitty')
    set_flag('mailman.django.purged')


# When we connect to the database units, request our database and extensions
@when_not('mailman.database.created')
@when('db.connected')
def request_db(pgsql):
    pgsql.set_database('mailman3')
    set_flag('mailman.database.created')


# Fetch the DB relation data and shove it into kv().
@when('db.master.available')
@when_not('mailman.database.configured')
def configure_database(pgsql):
    db_relation_config = pgsql.master
    db = unitdata.kv()
    db.set('db_name', db_relation_config.dbname)
    db.set('db_host', db_relation_config.host)
    db.set('db_user', db_relation_config.user)
    db.set('db_password', db_relation_config.password)
    write_config()
    set_flag('mailman.database.configured')


# Once the DB is full available, write the config.
@when('db.master.available')
@when('config.changed')
def write_updated_config():
    hookenv.status_set('maintenance', 'updating configuration')
    clear_flag('mailman.postfix.configured')
    write_config()
    hookenv.status_set('active', 'ready')


# Make sure we've got our REST API creds, then publish our relation data.
# This is a tiny bit odd, the relation.joined state is only cleared when the
# last related unit leaves the relation, but when it's initially set, i.e. on
# the relation actually being joined by the first related unit, it doesn't have
# any data.  So we guard the reaction with a relation.configured flag, it runs
# once on join with no data, the guard flag gets set, then when the relation
# changes, the interface sets relation.connected and we clear the guard flag
# so this can re-run.  In this case we're providing the relation, not consuming
# so it's slightly less relevant, but we still need to follow the same pattern.
@when('mailman.initial_setup')
@when('endpoint.mailman3.joined')
@when_not('mailman3.relation.configured')
def configure_mailman3():
    db = unitdata.kv()
    port = db.get('admin_port')
    ip = db.get('admin_ip')
    user = db.get('admin_user')
    password = db.get('admin_password')
    mailman3 = endpoint_from_flag('endpoint.mailman3.joined')
    mailman3.publish_info(ip, port, user, password)
    set_flag('mailman3.relation.configured')


# When the last unit leaves the relation, clear the guard flag so if/when
# a unit rejoins, we re-publish our info.
@when_not('endpoint.mailman3.joined')
def clear_mailman3_relation():
    clear_flag('mailman3.relation.configured')


# This state is the one set by the interface when any of the relation data
# changes, if that happens we clear the guard flag so we re-publish.
@when('endpoint.mailman3.connected')
def update_mailman3_relation():
    clear_flag('mailman3.relation.configured')


# This works similarly the mailman3 relation above except we're consuming the
# relation rather than publishing it, so it's even more important that we
# re-trigger configuration when the data changes, i.e. once we have data.
@when('endpoint.mailman3archiver.joined')
@when_not('mailman3archiver.relation.configured')
def configure_mailman3archiver():
    db = unitdata.kv()
    archivers = endpoint_from_flag('endpoint.mailman3archiver.joined')
    mailman3_archiver = archivers.mailman3()[0]
    db.set('archiver_ip', mailman3_archiver['ip'])
    db.set('archiver_key', mailman3_archiver['password'])
    write_config()
    set_flag('mailman3archiver.relation.configured')


@when_not('endpoint.mailman3archiver.joined')
def clear_archiver_relation():
    clear_flag('mailman3archiver.relation.configured')


@when_any('endpoint.mailman3archiver.changed.ip', 'endpoint.mailman3archiver.changed.password')
def changed_archiver_data():
    clear_flag('mailman3archiver.relation.configured')


@when('nrpe-external-master.available')
def setup_nagios(nagios):
    if hookenv.hook_name() == 'update-status':
        return

    hookenv.status_set('maintenance', 'setting up NRPE checks')
    # Ask charmhelpers.contrib.charmsupport's nrpe to work out our hostname
    hostname = nrpe.get_nagios_hostname()
    nrpe_setup = nrpe.NRPE(hostname=hostname)

    db = unitdata.kv()

    check_base = '/usr/lib/nagios/plugins/'
    process_check = check_base + 'check_procs'

    mailman_master_process = '{} -c 1:1 -C python3 -a master'.format(process_check)
    nrpe_setup.add_check('mailman_master_process', 'Check for Mailman3 master process', mailman_master_process)

    mailman_runner_process = '{} -c 12:12 -C python3 -a runner'.format(process_check)
    nrpe_setup.add_check('mailman_runner_process', 'Check for Mailman3 runner processes', mailman_runner_process)

    mailman_rest_port = '{}check_tcp -H {} -p {}'.format(check_base, db.get('admin_ip'), db.get('admin_port'))
    nrpe_setup.add_check('mailman_rest_port', 'Check Mailman3 REST API port', mailman_rest_port)

    postfix_check = '{}check_tcp -H 127.0.0.1 -p 25'.format(check_base)
    nrpe_setup.add_check('postfix_port', 'Check postfix port', postfix_check)

    nrpe_setup.write()
    hookenv.status_set('active', 'ready')


def systemctl(action, service_name):
    return subprocess.check_call(["systemctl", action, service_name])


def useradd(username):
    cmd = ["useradd",
           "-r",
           "-k", "/dev/null",
           "-md", "/var/lib/spamd",
           "-s", "/bin/false",
           username]
    try:
        pwd.getpwnam(username)
    except KeyError:
        return subprocess.check_call(cmd)


def postconf(key, value=None, args=None):
    cmd = ['postconf', '{k}'.format(k=key)]
    if value:
        cmd = ['postconf', '{k}={v}'.format(k=key, v=value)]
    if args:
        [cmd.insert(1, a) for a in reversed(args)]
    return subprocess.check_output(cmd)


def append_to_file(filep, content):
    with open(filep, "a") as f:
        f.write(content)
